@@ -42,10 +42,53 @@
 #define mpp(s,n) do { char *p = mpz_get_str(NULL,16,n); printf("%s 0x%sL\n", s, p); free(p); } while (0)
 
 struct rsacommon {
-    struct hash_if *hashi;
     uint8_t *hashbuf;
 };
 
+#define FREE(b)                ({ free((b)); (b)=0; })
+
+struct load_ctx {
+    void (*verror)(struct load_ctx *l, struct cloc loc,
+		   FILE *maybe_f,
+		   const char *message, va_list args);
+    bool_t (*postreadcheck)(struct load_ctx *l, FILE *f);
+    const char *what;
+    dict_t *deprdict; /* used only to look up hash */
+    struct cloc loc;
+    union {
+	struct {
+	    struct log_if *log;
+	} tryload;
+    } u;
+};
+
+static void load_err(struct load_ctx *l,
+		     const struct cloc *maybe_loc, FILE *maybe_f,
+		     const char *fmt, ...)
+{
+    va_list al;
+    va_start(al,fmt);
+    l->verror(l, maybe_loc ? *maybe_loc : l->loc, maybe_f,fmt,al);
+    va_end(al);
+}
+
+FORMAT(printf,4,0)
+static void verror_tryload(struct load_ctx *l, struct cloc loc,
+			   FILE *maybe_f,
+			   const char *message, va_list args)
+{
+    int class=M_ERR;
+    slilog_part(l->u.tryload.log,class,"%s: ",l->what);
+    vslilog(l->u.tryload.log,class,message,args);
+}
+
+static void verror_cfgfatal(struct load_ctx *l, struct cloc loc,
+			    FILE *maybe_f,
+			    const char *message, va_list args)
+{
+    vcfgfatal_maybefile(maybe_f,l->loc,l->what,message,args,"");
+}
+
 struct rsapriv {
     closure_t cl;
     struct sigprivkey_if ops;
@@ -56,6 +99,17 @@ struct rsapriv {
     MP_INT q, dq;
     MP_INT w;
 };
+
+#define RSAPUB_BNS(each)			\
+    each(0,e,"public exponent")			\
+    each(1,n,"modulus")
+
+#define RSAPUB_LOADCORE_PASSBN(ix,en,what) \
+    en##s, en##_loc,
+
+#define RSAPUB_INIT_ST_BN( ix,en,what) mpz_init (&st->en);
+#define RSAPUB_CLEAR_ST_BN(ix,en,what) mpz_clear(&st->en);
+
 struct rsapub {
     closure_t cl;
     struct sigpubkey_if ops;
@@ -74,25 +128,22 @@ struct rsapub {
 
 static const char *hexchars="0123456789abcdef";
 
-static void rsa_sethash(struct rsacommon *c, struct hash_if *hash)
+static void rsa_sethash(struct load_ctx *l,
+			struct rsacommon *c,
+			const struct hash_if **in_ops)
+{
+    struct hash_if *hash=0;
+    if (l->deprdict)
+	hash=find_cl_if(l->deprdict,"hash",CL_HASH,False,"site",l->loc);
+    if (!hash)
+	hash=sha1_hash_if;
+    c->hashbuf=safe_malloc(hash->hlen, "generate_msg");
+    *in_ops=hash;
+}
+
+static void rsacommon_dispose(struct rsacommon *c)
 {
     free(c->hashbuf);
-    c->hashbuf=safe_malloc(hash->hlen, "generate_msg");
-    c->hashi=hash;
-}
-static void rsa_pub_sethash(void *sst, struct hash_if *hash)
-{
-    struct rsapub *st=sst;
-    rsa_sethash(&st->common, hash);
-}
-static void rsa_priv_sethash(void *sst, struct hash_if *hash)
-{
-    struct rsapriv *st=sst;
-    rsa_sethash(&st->common, hash);
-}
-static void rsa_hash(struct rsacommon *c, const uint8_t *buf, int32_t len)
-{
-    hash_hash(c->hashi,buf,len,c->hashbuf);
 }
 
 static void emsa_pkcs1(MP_INT *n, MP_INT *m,
@@ -122,7 +173,7 @@ static void emsa_pkcs1(MP_INT *n, MP_INT *m,
     msize=mpz_sizeinbase(n, 16);
 
     if (datalen*2+6>=msize) {
-	fatal("rsa_sign: message too big");
+	fatal("rsa: message too big");
     }
 
     strcpy(buff,"0001");
@@ -154,9 +205,9 @@ static bool_t rsa_sign(void *sst, uint8_t *data, int32_t datalen,
     mpz_init(&a);
     mpz_init(&b);
 
-    rsa_hash(&st->common,data,datalen);
+    hash_hash(st->ops.hash,data,datalen,st->common.hashbuf);
     /* Construct the message representative. */
-    emsa_pkcs1(&st->n, &a, st->common.hashbuf, st->common.hashi->hlen);
+    emsa_pkcs1(&st->n, &a, st->common.hashbuf, st->ops.hash->hlen);
 
     /*
      * Produce an RSA signature (a^d mod n) using the Chinese
@@ -242,8 +293,8 @@ static bool_t rsa_sig_check(void *sst, uint8_t *data, int32_t datalen,
     mpz_init(&b);
     mpz_init(&c);
 
-    rsa_hash(&st->common,data,datalen);
-    emsa_pkcs1(&st->n, &a, st->common.hashbuf, st->common.hashi->hlen);
+    hash_hash(st->ops.hash,data,datalen,st->common.hashbuf);
+    emsa_pkcs1(&st->n, &a, st->common.hashbuf, st->ops.hash->hlen);
 
     /* Terminate signature with a '0' - already checked that this will fit */
     int save = sig->start[sig->len];
@@ -262,12 +313,27 @@ static bool_t rsa_sig_check(void *sst, uint8_t *data, int32_t datalen,
     return ok;
 }
 
-static list_t *rsapub_apply(closure_t *self, struct cloc loc, dict_t *context,
-			    list_t *args)
+static void rsapub_dispose(void *sst) {
+    struct rsapub *st=sst;
+
+    if (!st) return;
+    RSAPUB_BNS(RSAPUB_CLEAR_ST_BN)
+    rsacommon_dispose(&st->common);
+    free(st);
+}
+
+#define RSAPUB_LOADCORE_DEFBN(ix,en,what) \
+    const char *en##s, struct cloc en##_loc,
+
+#define LDPUBFATAL(lc,...) ({			\
+	load_err(l,(lc),0,__VA_ARGS__);	\
+	goto error_out;				\
+    })
+
+static struct rsapub *rsa_loadpub_core(RSAPUB_BNS(RSAPUB_LOADCORE_DEFBN)
+				       struct load_ctx *l)
 {
     struct rsapub *st;
-    item_t *i;
-    string_t e,n;
 
     NEW(st);
     st->cl.description="rsapub";
@@ -275,251 +341,304 @@ static list_t *rsapub_apply(closure_t *self, struct cloc loc, dict_t *context,
     st->cl.apply=NULL;
     st->cl.interface=&st->ops;
     st->ops.st=st;
-    st->ops.sethash=rsa_pub_sethash;
     st->common.hashbuf=NULL;
     st->ops.unpick=rsa_sig_unpick;
     st->ops.check=rsa_sig_check;
-    st->loc=loc;
+    st->ops.hash=0;
+    st->ops.dispose=rsapub_dispose;
+    st->loc=l->loc;
+    RSAPUB_BNS(RSAPUB_INIT_ST_BN)
 
-    i=list_elem(args,0);
-    if (i) {
-	if (i->type!=t_string) {
-	    cfgfatal(i->loc,"rsa-public","first argument must be a string\n");
-	}
-	e=i->data.string;
-	if (mpz_init_set_str(&st->e,e,10)!=0) {
-	    cfgfatal(i->loc,"rsa-public","encryption key \"%s\" is not a "
-		     "decimal number string\n",e);
-	}
-    } else {
-	cfgfatal(loc,"rsa-public","you must provide an encryption key\n");
-    }
-    if (mpz_sizeinbase(&st->e, 256) > RSA_MAX_MODBYTES) {
-	cfgfatal(loc, "rsa-public", "implausibly large public exponent\n");
-    }
-    
-    i=list_elem(args,1);
-    if (i) {
-	if (i->type!=t_string) {
-	    cfgfatal(i->loc,"rsa-public","second argument must be a string\n");
-	}
-	n=i->data.string;
-	if (mpz_init_set_str(&st->n,n,10)!=0) {
-	    cfgfatal(i->loc,"rsa-public","modulus \"%s\" is not a decimal "
-		     "number string\n",n);
-	}
-    } else {
-	cfgfatal(loc,"rsa-public","you must provide a modulus\n");
-    }
-    if (mpz_sizeinbase(&st->n, 256) > RSA_MAX_MODBYTES) {
-	cfgfatal(loc, "rsa-public", "implausibly large modulus\n");
+#define RSAPUB_LOADCORE_GETBN(ix,en,what)				\
+    if (mpz_init_set_str(&st->en,en##s,10)!=0) {			\
+	LDPUBFATAL(&en##_loc, what " \"%s\" is not a "			\
+		 "decimal number string",en##s);			\
+    }									\
+    if (mpz_sizeinbase(&st->en, 256) > RSA_MAX_MODBYTES) {		\
+	LDPUBFATAL(&en##_loc, "implausibly large " what);		\
     }
+
+    RSAPUB_BNS(RSAPUB_LOADCORE_GETBN)
+
+    rsa_sethash(l,&st->common,&st->ops.hash);
+
+    return st;
+
+ error_out:
+    rsapub_dispose(st);
+    return 0;
+}
+
+static list_t *rsapub_apply(closure_t *self, struct cloc loc, dict_t *context,
+			    list_t *args)
+{
+    struct load_ctx l[1];
+    l->verror=verror_cfgfatal;
+    l->postreadcheck=0;
+    l->what="rsa-public";
+    l->deprdict=context;
+    l->loc=loc;
+
+#define RSAPUB_APPLY_GETBN(ix,en,what)				\
+    item_t *en##i;						\
+    const char *en##s;						\
+    en##i=list_elem(args,ix);					\
+    if (!en##i)							\
+        cfgfatal(loc,"rsa-public",				\
+                 "you must provide an encryption key\n");	\
+    struct cloc en##_loc=en##i->loc;				\
+    if (en##i->type!=t_string)					\
+	cfgfatal(en##_loc,"rsa-public",				\
+		 "first argument must be a string\n");		\
+    en##s=en##i->data.string;
+
+    RSAPUB_BNS(RSAPUB_APPLY_GETBN)
+
+    struct rsapub *st=rsa_loadpub_core(RSAPUB_BNS(RSAPUB_LOADCORE_PASSBN)
+				       l);
+
     return new_closure(&st->cl);
 }
 
-static uint32_t keyfile_get_int(struct cloc loc, FILE *f)
+bool_t rsa1_loadpub(const struct sigscheme_info *algo,
+		    struct buffer_if *pubkeydata,
+		    struct sigpubkey_if **sigpub_r,
+		    closure_t **closure_r,
+		    struct log_if *log, struct cloc loc)
+{
+    struct rsapub *st=0;
+
+    struct load_ctx l[1];
+    l->verror=verror_tryload;
+    l->postreadcheck=0;
+    l->what="rsa1_loadpub";
+    l->deprdict=0;
+    l->loc=loc;
+    l->u.tryload.log=log;
+
+    char *nul=buf_append(pubkeydata,1);
+    if (!nul) LDPUBFATAL(0,"rsa1 public key data too long for extra nul");
+    *nul=0;
+
+    const char *delim=" \t\n";
+    char *saveptr;
+    /*unused*/ strtok_r(pubkeydata->start,delim,&saveptr);
+
+#define RSAPUB_TRYLOAD_GETBN(ix,en,what)				\
+    struct cloc en##_loc=loc;						\
+    const char *en##s=strtok_r(0,delim,&saveptr);			\
+    if (!en##s) LDPUBFATAL(0,"end of pubkey data looking for " what);
+
+    RSAPUB_BNS(RSAPUB_TRYLOAD_GETBN);
+
+    st=rsa_loadpub_core(RSAPUB_BNS(RSAPUB_LOADCORE_PASSBN) l);
+    if (!st) goto error_out;
+
+    *sigpub_r=&st->ops;
+    *closure_r=&st->cl;
+    return True;
+
+ error_out:
+    rsapub_dispose(st);
+    return False;
+}
+
+#define LDFATAL(...)      ({ load_err(l,0,0,__VA_ARGS__); goto error_out; })
+#define LDFATAL_FILE(...) ({ load_err(l,0,f,__VA_ARGS__); goto error_out; })
+#define KEYFILE_GET(is)   ({					\
+	uint##is##_t keyfile_get_tmp=keyfile_get_##is(l,f);	\
+	if (!l->postreadcheck(l,f)) goto error_out;		\
+	keyfile_get_tmp;					\
+    })
+
+static uint32_t keyfile_get_32(struct load_ctx *l, FILE *f)
 {
     uint32_t r;
     r=fgetc(f)<<24;
     r|=fgetc(f)<<16;
     r|=fgetc(f)<<8;
     r|=fgetc(f);
-    cfgfile_postreadcheck(loc,f);
     return r;
 }
 
-static uint16_t keyfile_get_short(struct cloc loc, FILE *f)
+static uint16_t keyfile_get_16(struct load_ctx *l, FILE *f)
 {
     uint16_t r;
     r=fgetc(f)<<8;
     r|=fgetc(f);
-    cfgfile_postreadcheck(loc,f);
     return r;
 }
 
-static list_t *rsapriv_apply(closure_t *self, struct cloc loc, dict_t *context,
-			     list_t *args)
+static void rsapriv_dispose(void *sst)
 {
-    struct rsapriv *st;
-    FILE *f;
-    cstring_t filename;
-    item_t *i;
+    struct rsapriv *st=sst;
+    mpz_clear(&st->n);
+    mpz_clear(&st->p); mpz_clear(&st->dp);
+    mpz_clear(&st->q); mpz_clear(&st->dq);
+    mpz_clear(&st->w);
+    rsacommon_dispose(&st->common);
+    free(st);
+}
+
+static struct rsapriv *rsa_loadpriv_core(struct load_ctx *l,
+					 FILE *f, struct cloc loc,
+					 bool_t do_validity_check)
+{
+    struct rsapriv *st=0;
     long length;
-    uint8_t *b, *c;
+    uint8_t *b=0, *c=0;
     int cipher_type;
     MP_INT e,d,iqmp,tmp,tmp2,tmp3;
     bool_t valid;
 
+    mpz_init(&e);
+    mpz_init(&d);
+    mpz_init(&iqmp);
+    mpz_init(&tmp);
+    mpz_init(&tmp2);
+    mpz_init(&tmp3);
+
     NEW(st);
     st->cl.description="rsapriv";
     st->cl.type=CL_SIGPRIVKEY;
     st->cl.apply=NULL;
     st->cl.interface=&st->ops;
     st->ops.st=st;
-    st->ops.sethash=rsa_priv_sethash;
     st->common.hashbuf=NULL;
     st->ops.sign=rsa_sign;
+    st->ops.hash=0;
+    st->ops.dispose=rsapriv_dispose;
     st->loc=loc;
+    mpz_init(&st->n);
+    mpz_init(&st->q);
+    mpz_init(&st->p);
+    mpz_init(&st->dp);
+    mpz_init(&st->dq);
+    mpz_init(&st->w);
 
-    /* Argument is filename pointing to SSH1 private key file */
-    i=list_elem(args,0);
-    if (i) {
-	if (i->type!=t_string) {
-	    cfgfatal(i->loc,"rsa-private","first argument must be a string\n");
-	}
-	filename=i->data.string;
-    } else {
-	filename=NULL; /* Make compiler happy */
-	cfgfatal(loc,"rsa-private","you must provide a filename\n");
-    }
-
-    f=fopen(filename,"rb");
     if (!f) {
-	if (just_check_config) {
-	    Message(M_WARNING,"rsa-private (%s:%d): cannot open keyfile "
-		    "\"%s\"; assuming it's valid while we check the "
-		    "rest of the configuration\n",loc.file,loc.line,filename);
-	    goto assume_valid;
-	} else {
-	    fatal_perror("rsa-private (%s:%d): cannot open file \"%s\"",
-			 loc.file,loc.line,filename);
-	}
+	assert(just_check_config);
+	goto assume_valid;
     }
 
     /* Check that the ID string is correct */
     length=strlen(AUTHFILE_ID_STRING)+1;
     b=safe_malloc(length,"rsapriv_apply");
     if (fread(b,length,1,f)!=1 || memcmp(b,AUTHFILE_ID_STRING,length)!=0) {
-	cfgfatal_maybefile(f,loc,"rsa-private","failed to read magic ID"
-			   " string from SSH1 private keyfile \"%s\"\n",
-			   filename);
+	LDFATAL_FILE("failed to read magic ID"
+		     " string from SSH1 private keyfile\n");
     }
-    free(b);
+    FREE(b);
 
     cipher_type=fgetc(f);
-    keyfile_get_int(loc,f); /* "Reserved data" */
+    KEYFILE_GET(32); /* "Reserved data" */
     if (cipher_type != 0) {
-	cfgfatal(loc,"rsa-private","we don't support encrypted keyfiles\n");
+	LDFATAL("we don't support encrypted keyfiles\n");
     }
 
     /* Read the public key */
-    keyfile_get_int(loc,f); /* Not sure what this is */
-    length=(keyfile_get_short(loc,f)+7)/8;
+    KEYFILE_GET(32); /* Not sure what this is */
+    length=(KEYFILE_GET(16)+7)/8;
     if (length>RSA_MAX_MODBYTES) {
-	cfgfatal(loc,"rsa-private","implausible length %ld for modulus\n",
+	LDFATAL("implausible length %ld for modulus\n",
 		 length);
     }
     b=safe_malloc(length,"rsapriv_apply");
     if (fread(b,length,1,f) != 1) {
-	cfgfatal_maybefile(f,loc,"rsa-private","error reading modulus\n");
+	LDFATAL_FILE("error reading modulus\n");
     }
-    mpz_init(&st->n);
     read_mpbin(&st->n,b,length);
-    free(b);
-    length=(keyfile_get_short(loc,f)+7)/8;
+    FREE(b);
+    length=(KEYFILE_GET(16)+7)/8;
     if (length>RSA_MAX_MODBYTES) {
-	cfgfatal(loc,"rsa-private","implausible length %ld for e\n",length);
+	LDFATAL("implausible length %ld for e\n",length);
     }
     b=safe_malloc(length,"rsapriv_apply");
     if (fread(b,length,1,f)!=1) {
-	cfgfatal_maybefile(f,loc,"rsa-private","error reading e\n");
+	LDFATAL_FILE("error reading e\n");
     }
-    mpz_init(&e);
     read_mpbin(&e,b,length);
-    free(b);
+    FREE(b);
     
-    length=keyfile_get_int(loc,f);
+    length=KEYFILE_GET(32);
     if (length>1024) {
-	cfgfatal(loc,"rsa-private","implausibly long (%ld) key comment\n",
+	LDFATAL("implausibly long (%ld) key comment\n",
 		 length);
     }
     c=safe_malloc(length+1,"rsapriv_apply");
     if (fread(c,length,1,f)!=1) {
-	cfgfatal_maybefile(f,loc,"rsa-private","error reading key comment\n");
+	LDFATAL_FILE("error reading key comment\n");
     }
     c[length]=0;
 
     /* Check that the next two pairs of characters are identical - the
        keyfile is not encrypted, so they should be */
 
-    if (keyfile_get_short(loc,f) != keyfile_get_short(loc,f)) {
-	cfgfatal(loc,"rsa-private","corrupt keyfile\n");
+    if (KEYFILE_GET(16) != KEYFILE_GET(16)) {
+	LDFATAL("corrupt keyfile\n");
     }
 
     /* Read d */
-    length=(keyfile_get_short(loc,f)+7)/8;
+    length=(KEYFILE_GET(16)+7)/8;
     if (length>RSA_MAX_MODBYTES) {
-	cfgfatal(loc,"rsa-private","implausibly long (%ld) decryption key\n",
+	LDFATAL("implausibly long (%ld) decryption key\n",
 		 length);
     }
     b=safe_malloc(length,"rsapriv_apply");
     if (fread(b,length,1,f)!=1) {
-	cfgfatal_maybefile(f,loc,"rsa-private",
-			   "error reading decryption key\n");
+	LDFATAL_FILE("error reading decryption key\n");
     }
-    mpz_init(&d);
     read_mpbin(&d,b,length);
-    free(b);
+    FREE(b);
     /* Read iqmp (inverse of q mod p) */
-    length=(keyfile_get_short(loc,f)+7)/8;
+    length=(KEYFILE_GET(16)+7)/8;
     if (length>RSA_MAX_MODBYTES) {
-	cfgfatal(loc,"rsa-private","implausibly long (%ld)"
+	LDFATAL("implausibly long (%ld)"
 		 " iqmp auxiliary value\n", length);
     }
     b=safe_malloc(length,"rsapriv_apply");
     if (fread(b,length,1,f)!=1) {
-	cfgfatal_maybefile(f,loc,"rsa-private",
-			   "error reading decryption key\n");
+	LDFATAL_FILE("error reading decryption key\n");
     }
-    mpz_init(&iqmp);
     read_mpbin(&iqmp,b,length);
-    free(b);
+    FREE(b);
     /* Read q (the smaller of the two primes) */
-    length=(keyfile_get_short(loc,f)+7)/8;
+    length=(KEYFILE_GET(16)+7)/8;
     if (length>RSA_MAX_MODBYTES) {
-	cfgfatal(loc,"rsa-private","implausibly long (%ld) q value\n",
+	LDFATAL("implausibly long (%ld) q value\n",
 		 length);
     }
     b=safe_malloc(length,"rsapriv_apply");
     if (fread(b,length,1,f)!=1) {
-	cfgfatal_maybefile(f,loc,"rsa-private",
-			   "error reading q value\n");
+	LDFATAL_FILE("error reading q value\n");
     }
-    mpz_init(&st->q);
     read_mpbin(&st->q,b,length);
-    free(b);
+    FREE(b);
     /* Read p (the larger of the two primes) */
-    length=(keyfile_get_short(loc,f)+7)/8;
+    length=(KEYFILE_GET(16)+7)/8;
     if (length>RSA_MAX_MODBYTES) {
-	cfgfatal(loc,"rsa-private","implausibly long (%ld) p value\n",
+	LDFATAL("implausibly long (%ld) p value\n",
 		 length);
     }
     b=safe_malloc(length,"rsapriv_apply");
     if (fread(b,length,1,f)!=1) {
-	cfgfatal_maybefile(f,loc,"rsa-private",
-			   "error reading p value\n");
+	LDFATAL_FILE("error reading p value\n");
     }
-    mpz_init(&st->p);
     read_mpbin(&st->p,b,length);
-    free(b);
+    FREE(b);
     
-    if (fclose(f)!=0) {
-	fatal_perror("rsa-private (%s:%d): fclose",loc.file,loc.line);
+    if (ferror(f)) {
+	fatal_perror("rsa-private (%s:%d): ferror",loc.file,loc.line);
     }
 
+    rsa_sethash(l,&st->common,&st->ops.hash);
+
     /*
      * Now verify the validity of the key, and set up the auxiliary
      * values for fast CRT signing.
      */
     valid=False;
-    i=list_elem(args,1);
-    mpz_init(&tmp);
-    mpz_init(&tmp2);
-    mpz_init(&tmp3);
-    if (i && i->type==t_bool && i->data.bool==False) {
-	Message(M_INFO,"rsa-private (%s:%d): skipping RSA key validity "
-		"check\n",loc.file,loc.line);
-    } else {
+    if (do_validity_check) {
 	/* Verify that p*q is equal to n. */
 	mpz_mul(&tmp, &st->p, &st->q);
 	if (mpz_cmp(&tmp, &st->n) != 0)
@@ -559,9 +678,6 @@ static list_t *rsapriv_apply(closure_t *self, struct cloc loc, dict_t *context,
      *   dq == d mod (q-1)      similarly mod q
      *   w == iqmp * q          so that w == 0 mod q, and w == 1 mod p
      */
-    mpz_init(&st->dp);
-    mpz_init(&st->dq);
-    mpz_init(&st->w);
     mpz_sub_ui(&tmp, &st->p, 1);
     mpz_mod(&st->dp, &d, &tmp);
     mpz_sub_ui(&tmp, &st->q, 1);
@@ -570,19 +686,131 @@ static list_t *rsapriv_apply(closure_t *self, struct cloc loc, dict_t *context,
     
 done_checks:
     if (!valid) {
-	cfgfatal(loc,"rsa-private","file \"%s\" does not contain a "
-		 "valid RSA key!\n",filename);
+	LDFATAL("file does not contain a "
+		 "valid RSA key!\n");
     }
+
+assume_valid:
+out:
     mpz_clear(&tmp);
     mpz_clear(&tmp2);
     mpz_clear(&tmp3);
 
-    free(c);
+    FREE(b);
+    FREE(c);
     mpz_clear(&e);
     mpz_clear(&d);
     mpz_clear(&iqmp);
 
-assume_valid:
+    return st;
+
+error_out:
+    if (st) rsapriv_dispose(st);
+    st=0;
+    goto out;
+}
+
+static bool_t postreadcheck_tryload(struct load_ctx *l, FILE *f)
+{
+    assert(!ferror(f));
+    if (feof(f)) { load_err(l,0,0,"eof mid-integer"); return False; }
+    return True;
+}
+
+bool_t rsa1_loadpriv(const struct sigscheme_info *algo,
+		     struct buffer_if *privkeydata,
+		     struct sigprivkey_if **sigpriv_r,
+		     closure_t **closure_r,
+		     struct log_if *log, struct cloc loc)
+{
+    FILE *f=0;
+    struct rsapriv *st=0;
+
+    f=fmemopen(privkeydata->start,privkeydata->size,"r");
+    if (!f) {
+	slilog(log,M_ERR,"failed to fmemopen private key file\n");
+	goto error_out;
+    }
+
+    struct load_ctx l[1];
+    l->what="rsa1priv load";
+    l->verror=verror_tryload;
+    l->postreadcheck=postreadcheck_tryload;
+    l->deprdict=0;
+    l->loc=loc;
+    l->u.tryload.log=log;
+
+    st=rsa_loadpriv_core(l,f,loc,False);
+    if (!st) goto error_out;
+    goto out;
+
+ error_out:
+    FREE(st);
+ out:
+    if (f) fclose(f);
+    if (!st) return False;
+    *sigpriv_r=&st->ops;
+    *closure_r=&st->cl;
+    return True;
+}
+
+static bool_t postreadcheck_apply(struct load_ctx *l, FILE *f)
+{
+    cfgfile_postreadcheck(l->loc,f);
+    return True;
+}
+
+static list_t *rsapriv_apply(closure_t *self, struct cloc loc, dict_t *context,
+			     list_t *args)
+{
+    struct rsapriv *st;
+    item_t *i;
+    cstring_t filename;
+    FILE *f;
+    struct load_ctx l[1];
+
+    l->what="rsa-private";
+    l->verror=verror_cfgfatal;
+    l->postreadcheck=postreadcheck_apply;
+    l->deprdict=context;
+    l->loc=loc;
+
+    /* Argument is filename pointing to SSH1 private key file */
+    i=list_elem(args,0);
+    if (i) {
+	if (i->type!=t_string) {
+	    cfgfatal(i->loc,"rsa-private","first argument must be a string\n");
+	}
+	filename=i->data.string;
+    } else {
+	filename=NULL; /* Make compiler happy */
+	cfgfatal(i->loc,"rsa-private","you must provide a filename\n");
+    }
+
+    f=fopen(filename,"rb");
+    if (!f) {
+	if (just_check_config) {
+	    Message(M_WARNING,"rsa-private (%s:%d): cannot open keyfile "
+		    "\"%s\" (%s); assuming it's valid while we check the "
+		    "rest of the configuration\n",
+		    loc.file,loc.line,filename,strerror(errno));
+	    return list_new();
+	} else {
+	    fatal_perror("rsa-private (%s:%d): cannot open file \"%s\"",
+			 loc.file,loc.line,filename);
+	}
+    }
+
+    bool_t do_validity_check=True;
+    i=list_elem(args,1);
+    if (i && i->type==t_bool && i->data.bool==False) {
+	Message(M_INFO,"rsa-private (%s:%d): skipping RSA key validity "
+		"check\n",loc.file,loc.line);
+	do_validity_check=False;
+    }
+
+    st=rsa_loadpriv_core(l,f,loc,do_validity_check);
+    fclose(f);
     return new_closure(&st->cl);
 }
 
